<?php
/**
 * @author      Lefteris Kavadas
 * @copyright   Copyright (c) 2016 - 2022 Lefteris Kavadas / firecoders.com
 * @license     GNU General Public License version 3 or later
 */
defined('_JEXEC') or die;

use Joomla\CMS\Filesystem\File;

require_once JPATH_SITE . '/plugins/system/route66pagespeed/lib/css.php';
require_once JPATH_SITE . '/plugins/system/route66pagespeed/lib/minifier.php';

class plgSystemRoute66PageSpeed extends JPlugin
{
	protected $doc;
	protected $cache;
	protected $optimizeJs = false;
	protected $optimizeCss = false;
	protected $optimizeImages = false;

	public function onAfterDispatch()
	{
		$application = JFactory::getApplication();
		$document = JFactory::getDocument();
		$user = JFactory::getUser();
		$componentParams = JComponentHelper::getParams('com_route66');

		if ($application->isClient('administrator') || $application->input->getMethod() != 'GET')
		{
			return;
		}

		if ($document->getType() != 'html')
		{
			return;
		}

		if (!$user->guest)
		{
			return;
		}

		if (in_array($application->input->getCmd('option'), $componentParams->get('optimizeExclusions', array())))
		{
			return;
		}

		$this->params->merge($componentParams);
		$this->optimizeJs = $this->params->get('optimizeJs');
		$this->optimizeCss = $this->params->get('optimizeCss');
		$this->optimizeImages = $this->params->get('lazyloadImages');
		$this->optimizeIframes = $this->params->get('lazyloadIframes') || $this->params->get('facadeIframes');
		$this->cache = JFactory::getCache('plg_system_route66_pagespeed', 'output');
		$this->cache->setCaching(true);
		$this->cache->setLifeTime($this->params->get('optimizeCacheTime', 15));

		if ($this->params->get('facadeIframes'))
		{
			$document->addStyleSheet(JUri::root(true) . '/media/route66/css/lazyframe.css');
			$document->addScript(JUri::root(true) . '/media/route66/js/lazyframe.min.js');
			$document->addScriptDeclaration('window.addEventListener("DOMContentLoaded", function(event) {lazyframe(".lazyframe");});');
		}
	}

	public function onAfterRender()
	{
		if ($this->optimizeJs || $this->optimizeCss || $this->optimizeImages)
		{
			$application = JFactory::getApplication();
			$buffer = '<?xml encoding="UTF-8">' . $application->getBody();

			libxml_use_internal_errors(true);
			$this->doc = new DOMDocument('1.0', 'UTF-8');
			$this->doc->preserveWhiteSpace = false;
			$this->doc->formatOutput = false;
			$this->doc->loadHTML($buffer);

			if ($this->optimizeImages)
			{
				$this->prepareImages();
			}

			if ($this->optimizeIframes)
			{
				$this->prepareIframes();
			}

			if ($this->optimizeJs)
			{
				$this->prepareJs();
			}

			if ($this->optimizeCss)
			{
				$this->prepareCss();
			}

			$buffer = $this->doc->saveHTML();
			$buffer = str_replace('<?xml encoding="UTF-8">', '', $buffer);

			$application->setBody($buffer);
		}
	}

	protected function prepareJs()
	{
		$scripts = $this->doc->getElementsByTagName('script');
		$parts = array();

		for ($i = $scripts->length; --$i >= 0;)
		{
			$script = $scripts->item($i);

			if (!$script->getAttribute('src'))
			{
				$type = $script->getAttribute('type');

				if ($type == 'text/javascript' || $type == '')
				{
					$parts[] = $script->nodeValue;
					$script->parentNode->removeChild($script);
				}
			}
			elseif (!$script->getAttribute('defer'))
			{
				$script->setAttribute('defer', 'defer');
			}
		}
		$js = implode('', array_reverse($parts));

		if ($js)
		{
			if ($this->params->get('optimizeJsMethod') == 'file')
			{
				$key = md5($js);
				$timestamp = $this->cache->get($key . '_script_timestamp');

				if (!$timestamp)
				{
					$date = JFactory::getDate();
					$timestamp = $date->toUnix();
					$this->cache->store($timestamp, $key . '_script_timestamp');
					File::write(JPATH_SITE . '/media/route66/scripts/' . $key . '.js', $js);
				}
				$head = $this->doc->getElementsByTagName('head')->item(0);
				$script = $this->doc->createElement('script', ' ');
				$script->setAttribute('defer', 'defer');
				$script->setAttribute('src', JUri::root(true) . '/media/route66/scripts/' . $key . '.js?v=' . $timestamp);
				$head->appendChild($script);
			}
			else
			{
				$body = $this->doc->getElementsByTagName('body')->item(0);
				$script = $this->doc->createElement('script', 'document.addEventListener("DOMContentLoaded", function(event) {' . $js . '});');
				$body->appendChild($script);
			}
		}
	}

	protected function prepareCss()
	{
		$xpath = new DOMXpath($this->doc);
		$styles = $xpath->query('//link[@rel="stylesheet"] | //style');

		$processor = new Route66Css();
		$processed = array();

		foreach ($styles as $key => $style)
		{
			if ($style->tagName == 'style')
			{
				$result = $processor->addStyle($style->nodeValue);
			}
			else
			{
				$result = $processor->addLink($style->getAttribute('href'));
			}

			if ($result)
			{
				$processed[] = $key;
			}
		}

		for ($i = $styles->length; --$i >= 0;)
		{
			$style = $styles->item($i);

			if (in_array($i, $processed))
			{
				$style->parentNode->removeChild($style);
			}
		}

		$css = $processor->combine();

		if ($css)
		{
			if ($this->params->get('minifyCss'))
			{
				$minifier = new Route66CssMinifier();
				$css = $minifier->process($css);
			}

			$head = $this->doc->getElementsByTagName('head')->item(0);
			$style = $this->doc->createElement('style', $css);
			$head->appendChild($style);
		}
	}

	protected function prepareImages()
	{
		$componentParams = JComponentHelper::getParams('com_route66');
		$restrictionMode = $componentParams->get('lazyloadImagesRestrictionsMode');
		$classname = $componentParams->get('lazyloadImagesClass');

		$images = $this->doc->getElementsByTagName('img');

		for ($i = $images->length; --$i >= 0;)
		{
			$image = $images->item($i);

			$class = $image->getAttribute('class');

			if ($restrictionMode && $classname)
			{
				$classes = explode(' ', $class);
				$classes = array_filter($classes);

				if ($restrictionMode == 'inclusive' && !in_array($classname, $classes))
				{
					continue;
				}
				elseif ($restrictionMode == 'exclusive' && in_array($classname, $classes))
				{
					continue;
				}
			}

			$image->setAttribute('loading', 'lazy');
		}
	}

	protected function prepareIframes()
	{
		$componentParams = JComponentHelper::getParams('com_route66');

		$lazyload = $componentParams->get('lazyloadIframes');
		$lazyloadRestrictionMode = $componentParams->get('lazyloadIframesRestrictionsMode');
		$lazyloadClassname = $componentParams->get('lazyloadIframesClass');

		$facades = $componentParams->get('facadeIframes');
		$facadesRestrictionMode = $componentParams->get('facadeIframesRestrictionsMode');
		$facadesClassname = $componentParams->get('facadeIframesClass');

		$iframes = $this->doc->getElementsByTagName('iframe');

		for ($i = $iframes->length; --$i >= 0;)
		{
			$iframe = $iframes->item($i);

			$class = $iframe->getAttribute('class');

			$classes = explode(' ', $class);
			$classes = array_filter($classes);

			if ($lazyload)
			{
				if (!$lazyloadRestrictionMode || ($lazyloadRestrictionMode === 'inclusive' && in_array($lazyloadClassname, $classes)) || ($lazyloadRestrictionMode === 'exclusive' && !in_array($lazyloadClassname, $classes)))
				{
					$iframe->setAttribute('loading', 'lazy');
				}
			}

			if ($facades)
			{
				if (!$facadesRestrictionMode || ($facadesRestrictionMode === 'inclusive' && in_array($facadesClassname, $classes)) || ($facadesRestrictionMode === 'exclusive' && !in_array($facadesClassname, $classes)))
				{
					$src = $iframe->getAttribute('src');
					$title = $iframe->getAttribute('title');
					$width = (int) $iframe->getAttribute('width');
					$height = (int) $iframe->getAttribute('height');

					$vendor = '';

					if (strpos($src, 'youtube.com/embed/') || strpos($src, 'youtube-nocookie.com/embed/'))
					{
						$vendor = strpos($src, 'youtube.com/embed/') ? 'youtube' : 'youtube_nocookie';
					}
					elseif (strpos($src, 'vimeo.com/video/'))
					{
						$vendor = 'vimeo';
					}

					if (!$vendor)
					{
						continue;
					}

					$class .= ' lazyframe';

					$div = $this->doc->createElement('div');
					$div->setAttribute('data-src', $src);
					$div->setAttribute('class', $class);

					if ($vendor)
					{
						$div->setAttribute('data-vendor', $vendor);
					}

					if ($title)
					{
						$div->setAttribute('data-title', $title);
					}

					if ($width && $height)
					{
						$div->setAttribute('style', 'aspect-ratio: ' . $width . ' / ' . $height . ';');
					}

					$iframe->parentNode->replaceChild($div, $iframe);
				}
			}
		}
	}
}
